close all;
clear all;

scale = 0.5; % scale of the input image 
k = 5; % number of colors 
alpha = 0.1;
method = 'PDHG' 
%method = 'ADMM'

% read image 
img = imread('247085.jpg');
im = double(imresize( img, scale )) / 255.0;
[ny,nx,~] = size(im);
n = ny * nx; % number of pixels

% compute colors using kmeans 
[~, C] = kmeans(reshape(im, [n, 3]), k);

% compute data term 
f = zeros(ny, nx, k);
for i=1:k
    f(:, :, i) = sqrt(sum((im - repmat(reshape(C(i, :), 1, 1, 3), ny, nx)).^2, 3));
end
f = reshape(f, [n, k])';

% set up gradient operator D
e = ones(n, 1);
Dx_tilde = spdiags([e -e], -1:0, nx, nx); 
Dx_tilde(:, end) = 0;
e = ones(nx, 1);
Dy_tilde = spdiags([-e e], 0:1, ny, ny);
Dy_tilde(end, end) = 0;
Dx = kron(Dx_tilde', speye(ny));
Dy = kron(speye(nx), Dy_tilde);
D = [Dx; Dy];

%% ADMM
if strcmp(method, 'ADMM')
    max_iters = 50;
    u = zeros(k, n);
    v1 = zeros(k, n);
    v2 = zeros(k, 2*n);
    q1 = zeros(k, n);
    q2 = zeros(k, 2*n);
    rho = 0.1;

    for it=1:max_iters
        it
        % primal update 1
        for i=1:k
            rhs = v1(i, :)' + D' * (v2(i, :)' - q2(i, :)' / rho) - (f(i, :)' + q1(i, :)') / rho;
            u(i, :) = (speye(n) + D' * D) \ rhs;
        end
    
        % primal update 2
        proj_arg = u + q1 / rho;
        v1 = projSimplex(proj_arg')';
    
        for i=1:k
            prox_arg = D*u(i, :)' + q2(i,:)' / rho;
            v2(i, :) = prox_l1(prox_arg, alpha / rho)';
        end
    
        % Lagrange multiplier update
        q1 = q1 + rho * (u - v1);
        for i=1:k
            q2(i, :) = q2(i, :)' + rho * (D * u(i, :)' - v2(i, :)');
        end
    end
end

%% PDHG
if strcmp(method, 'PDHG')
    u = zeros(k, n);
    u_tilde = zeros(k, n);
    q = zeros(k, 2*n);
    tau = 0.25;
    sigma = 0.5;
    max_iters=250;

    for it=1:max_iters
        for i=1:k
            u_tilde(i, :) = u(i, :)' - tau * (D' * q(i, :)' + f(i, :)');
        end
        u_prev = u;
        u = projSimplex(u_tilde')';
    
        for i=1:k
            q(i, :) = min(max( q(i,:)' + sigma * D * (2*u(i,:)' - u_prev(i,:)'), -alpha), alpha);
        end
    end
end

% visualize result 
[~, classes] = max(u, [], 1);
im_result = reshape(C(classes, :), [ny nx 3]);
imshow([im; im_result]);
